# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Oumi (Open Universal Machine Intelligence) library.

This library provides tools and utilities for training, evaluating, and
inferring with machine learning models, particularly focused on language tasks.

Modules:
    - :mod:`~oumi.models`: Contains model architectures and related utilities.
    - :mod:`~oumi.evaluate`: Functions for evaluating models.
    - :mod:`~oumi.evaluate_async`: Asynchronous evaluation functionality.
    - :mod:`~oumi.infer`: Functions for model inference, including interactive mode.
    - :mod:`~oumi.train`: Training utilities for machine learning models.
    - :mod:`~oumi.utils`: Utility functions, including logging configuration.
    - :mod:`~oumi.judges`: Functions for judging datasets and model responses.

Functions:
    - :func:`~oumi.train.train`: Train a machine learning model.
    - :func:`~oumi.evaluate_async.evaluate_async`: Asynchronously evaluate a model.
    - :func:`~oumi.evaluate.evaluate`: Evaluate a model using LM Harness.
    - :func:`~oumi.infer.infer`: Perform inference with a trained model.
    - :func:`~oumi.infer.infer_interactive`: Run interactive inference with a model.
    - :func:`~oumi.judge.judge_dataset`: Judge a dataset using a model.

Examples:
    Training a model::

        >>> from oumi import train
        >>> from oumi.core.configs import TrainingConfig
        >>> config = TrainingConfig(...)
        >>> train(config)

    Evaluating a model::

        >>> from oumi import evaluate
        >>> from oumi.core.configs import EvaluationConfig
        >>> config = EvaluationConfig(...)
        >>> results = evaluate(config)

    Performing inference::

        >>> from oumi import infer
        >>> from oumi.core.configs import InferenceConfig
        >>> config = InferenceConfig(...)
        >>> outputs = infer(config)

    Judging a dataset::

        >>> from oumi import judge_dataset
        >>> from oumi.core.configs import JudgeConfig
        >>> config = JudgeConfig(...)
        >>> judge_dataset(config, dataset)

See Also:
    - :mod:`oumi.core.configs`: For configuration classes used in Oumi
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from oumi.utils import logging

if TYPE_CHECKING:
    from oumi.core.configs import (
        AsyncEvaluationConfig,
        EvaluationConfig,
        InferenceConfig,
        JudgeConfig,
        TrainingConfig,
    )
    from oumi.core.datasets import BaseSftDataset
    from oumi.core.inference import BaseInferenceEngine
    from oumi.core.types.conversation import Conversation

logging.configure_dependency_warnings()


def evaluate_async(config: AsyncEvaluationConfig) -> None:
    """Runs an async evaluation for a model using the provided configuration.

    Overview:
        This is a utility method for running evaluations iteratively over a series
        of checkpoints. This method can be run in parallel with a training job to
        compute metrics per checkpoint without wasting valuable time in the main
        training loop.

    Args:
        config: The desired configuration for evaluation.

    Returns:
        None.
    """
    import oumi.evaluate_async

    return oumi.evaluate_async.evaluate_async(config)


def evaluate(config: EvaluationConfig) -> list[dict[str, Any]]:
    """Evaluates a model using the provided configuration.

    Args:
        config: The desired configuration for evaluation.

    Returns:
        A list of evaluation results (one for each task). Each evaluation result is a
        dictionary of metric names and their corresponding values.
    """
    import oumi.evaluate

    return oumi.evaluate.evaluate(config)


def infer_interactive(
    config: InferenceConfig,
    *,
    input_image_bytes: list[bytes] | None = None,
    system_prompt: str | None = None,
) -> None:
    """Interactively provide the model response for a user-provided input."""
    import oumi.infer

    return oumi.infer.infer_interactive(
        config, input_image_bytes=input_image_bytes, system_prompt=system_prompt
    )


def infer(
    config: InferenceConfig,
    inputs: list[str] | None = None,
    inference_engine: BaseInferenceEngine | None = None,
    *,
    input_image_bytes: list[bytes] | None = None,
) -> list[Conversation]:
    """Runs batch inference for a model using the provided configuration.

    Args:
        config: The configuration to use for inference.
        inputs: A list of inputs for inference.
        inference_engine: The engine to use for inference. If unspecified, the engine
            will be inferred from `config`.
        input_image_bytes: A list of input PNG image bytes to be used with `image+text`
            VLMs. Only used in interactive mode.

    Returns:
        object: A list of model responses.
    """
    import oumi.infer

    return oumi.infer.infer(
        config, inputs, inference_engine, input_image_bytes=input_image_bytes
    )


def judge_conversations(
    config: JudgeConfig, judge_inputs: list[Conversation]
) -> list[dict[str, Any]]:
    """Judge a list of conversations.

    This function evaluates a list of conversations using the specified Judge.

    The function performs the following steps:

        1. Initializes the Judge with the provided configuration.
        2. Uses the Judge to evaluate each conversation input.
        3. Collects and returns the judged outputs.

    Args:
        config: The configuration for the judge.
        judge_inputs: A list of Conversation objects to be judged.

    Returns:
        List[Dict[str, Any]]: A list of judgement results for each conversation.

        >>> # Example output:
        [
            {'helpful': True, 'safe': False},
            {'helpful': True, 'safe': True},
        ]

    Example:
        >>> config = JudgeConfig(...) # doctest: +SKIP
        >>> judge_inputs = [Conversation(...), Conversation(...)] # doctest: +SKIP
        >>> judged_outputs = judge_conversations(config, judge_inputs) # doctest: +SKIP
        >>> for output in judged_outputs: # doctest: +SKIP
        ...     print(output)
    """
    import oumi.judge

    return oumi.judge.judge_conversations(config, judge_inputs)


def judge_dataset(config: JudgeConfig, dataset: BaseSftDataset) -> list[dict[str, Any]]:
    """Judge a dataset.

    This function evaluates a given dataset using a specified Judge configuration.

    The function performs the following steps:

        1. Initializes the Judge with the provided configuration.
        2. Iterates through the dataset to extract conversation inputs.
        3. Uses the Judge to evaluate each conversation input.
        4. Collects and returns the judged outputs.

    Args:
        config: The configuration for the judge.
        dataset: The dataset to be judged. This dataset
            should be compatible with the Supervised Finetuning Dataset class.

    Returns:
        List[Dict[str, Any]]: A list of judgement results for each conversation.

        >>> # Example output:
        [
            {'helpful': True, 'safe': False},
            {'helpful': True, 'safe': True},
        ]

    Example:
        >>> config = JudgeConfig(...) # doctest: +SKIP
        >>> dataset = SomeDataset(...) # doctest: +SKIP
        >>> judged_outputs = judge_dataset(config, dataset) # doctest: +SKIP
        >>> for output in judged_outputs: # doctest: +SKIP
        ...     print(output)
    """
    import oumi.judge

    return oumi.judge.judge_dataset(config, dataset)


def train(config: TrainingConfig, **kwargs) -> None:
    """Trains a model using the provided configuration."""
    import oumi.train

    return oumi.train.train(config, *kwargs)


__all__ = [
    "evaluate_async",
    "evaluate",
    "infer_interactive",
    "infer",
    "judge_conversations",
    "judge_dataset",
    "train",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Final

import torch

from oumi.core.configs.base_config import BaseConfig
from oumi.core.configs.params.data_params import DataParams
from oumi.core.configs.params.fsdp_params import FSDPParams
from oumi.core.configs.params.model_params import ModelParams
from oumi.core.configs.params.peft_params import PeftParams
from oumi.core.configs.params.training_params import (
    MixedPrecisionDtype,
    TrainerType,
    TrainingParams,
)
from oumi.utils.logging import logger


@dataclass
class TrainingConfig(BaseConfig):
    data: DataParams = field(default_factory=DataParams)
    """Parameters for the dataset.

    This field contains all the necessary settings for data processing and loading.
    It includes options for train and evaluation datasets and preprocessing steps.

    For more details, see the :class:`oumi.core.configs.params.data_params.DataParams`
    class.
    """

    model: ModelParams = field(default_factory=ModelParams)
    """Parameters for the model.

    This field defines the model architecture, size, and other model-specific settings.
    It includes options for model type, pretrained weights, and tokenizer configuration.

    For more details, see :class:`oumi.core.configs.params.model_params.ModelParams`
    class.
    """

    training: TrainingParams = field(default_factory=TrainingParams)
    """Parameters for the training process.

    This field contains all settings related to the training loop,
    including learning rate, batch size, number of epochs, and optimization parameters.

    For more details, see
    :class:`oumi.core.configs.params.training_params.TrainingParams`.
    """

    peft: PeftParams = field(default_factory=PeftParams)
    """Parameters for Parameter-Efficient Fine-Tuning (PEFT).

    This field defines settings for various PEFT methods such as LoRA, or Prefix Tuning.
    It includes options for rank, alpha values, and other PEFT-specific parameters.

    For more details, see :class:`oumi.core.configs.params.peft_params.PeftParams`.
    """

    fsdp: FSDPParams = field(default_factory=FSDPParams)
    """Parameters for FSDP."""

    def __post_init__(self):
        """Verifies/populates params."""
        if self.model.compile:
            raise ValueError(
                "Use `training.compile` instead of `model.compile` to "
                "enable model compilation during training."
            )
        if self.training.compile and (
            self.fsdp.use_orig_params is not None and not self.fsdp.use_orig_params
        ):
            raise ValueError(
                "`fsdp.use_orig_params` must be True for model compilation."
            )

        # Verify values for model dtype and mixed precision training.
        if self.training.mixed_precision_dtype in [
            MixedPrecisionDtype.FP16,
            MixedPrecisionDtype.BF16,
        ]:
            if self.model.torch_dtype != torch.float32:
                raise ValueError(
                    "Model must be loaded in fp32 to enable mixed precision training."
                )

        trainer_type: Final[TrainerType] = self.training.trainer_type

        # Check values for model sequence length.
        if self.model.model_max_length and self.model.model_max_length > 0:
            max_seq_length_value = int(self.model.model_max_length)
            max_seq_length_key = None
            if trainer_type == TrainerType.TRL_SFT:
                max_seq_length_key = "max_seq_length"
            elif trainer_type == TrainerType.TRL_DPO:
                max_seq_length_key = "max_length"
                # TODO: DPOTrainer also defines "max_prompt_length" and
                # "max_target_length". How to handle them?
            else:
                logger.warning(
                    f"Ignored model.model_max_length={max_seq_length_value} "
                    f"parameter for trainer {self.training.trainer_type}."
                )

            if max_seq_length_key:
                existing_max_seq_length = self.training.trainer_kwargs.get(
                    max_seq_length_key
                )
                if (existing_max_seq_length is not None) and (
                    existing_max_seq_length != max_seq_length_value
                ):
                    logger.warning(
                        f"Overriding existing '{max_seq_length_key}' value "
                        f"'{existing_max_seq_length}' with '{max_seq_length_value}'"
                    )
                self.training.trainer_kwargs[max_seq_length_key] = max_seq_length_value

        # Set Liger kernel flags if using a HF trainer, and if so, don't do Liger
        # patch ourselves.
        # TODO(OPE-1117): Clean up this logic after upgrading to trl 0.16.
        if self.model.enable_liger_kernel:
            if trainer_type == TrainerType.TRL_SFT:
                self.training.trainer_kwargs["use_liger"] = True
                self.training.trainer_kwargs["use_liger_kernel"] = True
                self.model.enable_liger_kernel = False
            elif trainer_type in (TrainerType.TRL_DPO, TrainerType.HF):
                self.training.trainer_kwargs["use_liger_kernel"] = True
                self.model.enable_liger_kernel = False
            elif trainer_type == TrainerType.OUMI:
                # We need to Liger patch ourselves for our own training loop.
                pass
            else:
                raise ValueError("Unrecognized trainer type!")

        # Setup and validate params for "vision_language_sft" collator.
        # The collator expects VLM SFT dataset to only produce just
        # one column: 'conversation_json' (JSON-encoded `Conversation`)!
        collator_name: Final[str] = self.data.train.collator_name or ""
        if collator_name == "vision_language_sft":
            for dataset_params in self.data.train.datasets:
                if not dataset_params.dataset_kwargs.get("return_conversations", True):
                    raise ValueError(
                        "`return_conversations` must be True "
                        f"for the dataset '{dataset_params.dataset_name}' "
                        f"when using '{collator_name}' collator!"
                    )
                dataset_params.dataset_kwargs["return_conversations"] = True
            # Extra setup for TRL_SFT.
            if trainer_type == TrainerType.TRL_SFT:
                if self.training.trainer_kwargs.get("remove_unused_columns", False):
                    raise ValueError(
                        "`remove_unused_columns` must be False "
                        f"when using '{collator_name}' collator! "
                        'The "unused" columns are consumed by the collator, '
                        "not by a model."
                    )
                self.training.trainer_kwargs["remove_unused_columns"] = False

                # `trl` shouldn't be preparing the dataset, as we do it in Oumi.
                dataset_kwargs = self.training.trainer_kwargs.get("dataset_kwargs", {})
                dataset_kwargs["skip_prepare_dataset"] = True
                self.training.trainer_kwargs["dataset_kwargs"] = dataset_kwargs



from oumi.core.configs.params.remote_params import RemoteParams


def test_remote_params_allows_empty():
    params = RemoteParams()
    params.finalize_and_validate()
    # No exception should be raised



from unittest.mock import Mock, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import BaseCloud, BaseCluster, JobStatus
from oumi.launcher.launcher import (
    LAUNCHER,
    Launcher,
    cancel,
    down,
    get_cloud,
    run,
    status,
    stop,
    up,
    which_clouds,
)


#
# Fixtures
#
@pytest.fixture
def mock_registry():
    with patch("oumi.launcher.launcher.REGISTRY") as registry:
        yield registry


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_launcher_get_cloud(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
    }
    launcher = Launcher()
    cloud = launcher.get_cloud(_get_default_job("sky"))
    assert cloud == sky_mock
    assert cloud != polaris_mock


def test_launcher_get_cloud_missing_value(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        sky_mock = Mock(spec=BaseCloud)
        polaris_mock = Mock(spec=BaseCloud)

        def _sky_builder():
            return sky_mock

        def _polaris_builder():
            return polaris_mock

        mock_registry.get.return_value = None
        mock_registry.get_all.return_value = {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
        }
        launcher = Launcher()
        launcher.get_cloud(_get_default_job("lambda"))
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_empty(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_registry.get_all.return_value = {}
        mock_registry.get.return_value = None
        launcher = Launcher()
        launcher.get_cloud(_get_default_job("sky"))
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_by_name_missing_value(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        sky_mock = Mock(spec=BaseCloud)
        polaris_mock = Mock(spec=BaseCloud)

        def _sky_builder():
            return sky_mock

        def _polaris_builder():
            return polaris_mock

        mock_registry.get_all.return_value = {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
        }
        mock_registry.get.return_value = None
        launcher = Launcher()
        _ = launcher.get_cloud("lambda")
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_by_name_empty(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_registry.get_all.return_value = {}
        mock_registry.get.return_value = None
        launcher = Launcher()
        _ = launcher.get_cloud("lambda")
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_up_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, "custom")
    mock_cloud.up_cluster.assert_called_once_with(job, "custom")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_succeeds_kwargs(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, "custom", foo="bar")
    mock_cloud.up_cluster.assert_called_once_with(job, "custom", foo="bar")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_succeeds_no_name(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, None)
    mock_cloud.up_cluster.assert_called_once_with(job, None)
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_inavlid_cluster(mock_registry):
    with pytest.raises(RuntimeError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "custom": _builder,
        }
        expected_job_status = JobStatus(
            id="job_id",
            cluster="custom",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        )
        mock_cloud.up_cluster.return_value = expected_job_status
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        job = _get_default_job("custom")
        launcher.up(job, None)
    assert "not found" in str(exception_info.value)


def test_launcher_run_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
    )
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.run_job.return_value = expected_job_status
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.run(job, "custom")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    mock_cluster.run_job.assert_called_once_with(job)
    assert result == expected_job_status


def test_launcher_run_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "custom": _builder,
        }
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        job = _get_default_job("custom")
        launcher.run(job, "custom")
    assert "not found" in str(exception_info.value)


def test_launcher_cancel_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    expected_job_status = JobStatus(
        id="job_id",
        cluster="cluster",
        name="foo",
        status="canceled",
        metadata="bar",
        done=False,
    )
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.cancel_job.return_value = expected_job_status
    launcher = Launcher()
    result = launcher.cancel("1", "cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.cancel_job.assert_called_once_with("1")
    assert result == expected_job_status


def test_launcher_cancel_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.cancel("1", "cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_down_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    launcher.down("cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.down.assert_called_once()


def test_launcher_down_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.down("cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_status_multiple_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status()
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
    }


def test_launcher_status_filters_clusters(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.name.return_value = "sky1"
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.name.return_value = "sky2"
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.name.return_value = "polaris1"
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.name.return_value = "polaris2"
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.name.return_value = "polaris3"
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cluster="sky1")
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
        "polaris": [],
    }


def test_launcher_status_filters_jobs(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(id="1")
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
    }


def test_launcher_status_filters_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cloud="sky")
    assert statuses == {
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
    }


def test_launcher_status_all_filters(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.name.return_value = "foo"
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.name.return_value = "sky2"
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.name.return_value = "foo"
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.name.return_value = "polaris2"
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.name.return_value = "polaris3"
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cluster="foo", cloud="sky", id="1")
    assert statuses == {
        "sky": [
            JobStatus(
                id="1",
                cluster="foo",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
    }


def test_launcher_status_inits_new_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.side_effect = [
        {},
        {},
        {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
            "custom": _custom_builder,
        },
    ]
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status()
    # On the first call, statuses should be empty.
    assert statuses == {}
    # On the second call, we've registered new clouds that yield jobs.
    new_statuses = launcher.status()
    assert new_statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
            ),
        ],
    }


def test_launcher_stop_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    launcher.stop("cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.stop.assert_called_once()


def test_launcher_stop_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.stop("cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_which_clouds_updates_over_time(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.side_effect = [
        {},
        {
            "sky": _sky_builder,
        },
        {
            "polaris": _polaris_builder,
        },
        {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
            "custom": _custom_builder,
        },
    ]
    launcher = Launcher()
    assert launcher.which_clouds() == ["sky"]
    assert launcher.which_clouds() == ["polaris"]
    assert launcher.which_clouds() == ["sky", "polaris", "custom"]


def test_launcher_export_methods(mock_registry):
    assert LAUNCHER.up == up
    assert LAUNCHER.run == run
    assert LAUNCHER.cancel == cancel
    assert LAUNCHER.down == down
    assert LAUNCHER.status == status
    assert LAUNCHER.stop == stop
    assert LAUNCHER.get_cloud == get_cloud
    assert LAUNCHER.which_clouds == which_clouds



import tempfile
from pathlib import Path

import pytest
import typer
from typer.testing import CliRunner

from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.judge import conversations, dataset, model
from oumi.core.types import Conversation, Message
from oumi.core.types.conversation import Role
from oumi.utils.io_utils import save_jsonlines

runner = CliRunner()

config = "oumi/v1_xml_unit_test"


#
# Fixtures
#
@pytest.fixture
def app():
    judge_app = typer.Typer()
    judge_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(dataset)
    judge_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(conversations)
    judge_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(model)
    yield judge_app


def test_judge_dataset_runs(app):
    config = "oumi/v1_xml_unit_test"
    result = runner.invoke(
        app,
        [
            "dataset",
            "--config",
            config,
            "--dataset-name",
            "debug_sft",
        ],
    )

    assert result.exit_code == 0, f"CLI command failed with: {result.exception}"


def test_judge_dataset_with_output_file(app):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        output_file = str(Path(output_temp_dir) / "output.jsonl")

        result = runner.invoke(
            app,
            [
                "dataset",
                "--config",
                config,
                "--dataset-name",
                "debug_sft",
                "--output-file",
                output_file,
            ],
        )

        assert result.exit_code == 0
        assert Path(output_file).exists()


def test_judge_conversations_runs(app):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        input_file = str(Path(output_temp_dir) / "input.jsonl")

        conversation = Conversation(
            messages=[
                Message(role=Role.USER, content="Hello"),
                Message(role=Role.ASSISTANT, content="Hello"),
            ]
        )

        save_jsonlines(
            input_file,
            [conversation.to_dict()],
        )

        result = runner.invoke(
            app,
            [
                "conversations",
                "--config",
                config,
                "--input-file",
                input_file,
            ],
        )

        assert result.exit_code == 0


def test_judge_conversations_with_output_file(app):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        input_file = str(Path(output_temp_dir) / "input.jsonl")
        conversation = Conversation(
            messages=[
                Message(role=Role.USER, content="Hello"),
                Message(role=Role.ASSISTANT, content="Hello"),
            ]
        )

        save_jsonlines(
            input_file,
            [conversation.to_dict()],
        )

        output_file = str(Path(output_temp_dir) / "output.jsonl")

        result = runner.invoke(
            app,
            [
                "conversations",
                "--config",
                config,
                "--input-file",
                input_file,
                "--output-file",
                output_file,
            ],
        )

        assert result.exit_code == 0
        assert Path(output_file).exists()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from oumi.core.configs import JudgeConfig
from oumi.core.datasets import BaseSftDataset
from oumi.core.types.conversation import Conversation
from oumi.judges.oumi_judge import OumiXmlJudge as Judge


def judge_dataset(config: JudgeConfig, dataset: BaseSftDataset) -> list[dict[str, Any]]:
    """Judge a dataset.

    This function evaluates a given dataset using a specified Judge configuration.

    The function performs the following steps:

        1. Initializes the Judge with the provided configuration.
        2. Iterates through the dataset to extract conversation inputs.
        3. Uses the Judge to evaluate each conversation input.
        4. Collects and returns the judged outputs.

    Args:
        config: The configuration for the judge.
        dataset: The dataset to be judged. This dataset
            should be compatible with the Supervised Finetuning Dataset class.

    Returns:
        List[Dict[str, Any]]: A list of judgement results for each conversation.

        >>> # Example output:
        [
            {'helpful': True, 'safe': False},
            {'helpful': True, 'safe': True},
        ]

    Example:
        >>> config = JudgeConfig(...)
        >>> dataset = SomeDataset(...)
        >>> judged_outputs = judge_dataset(config, dataset)
        >>> for output in judged_outputs:
        ...     print(output)
    """
    judge = Judge(config)
    judge_inputs = [dataset.conversation(idx) for idx in range(len(dataset))]
    judge_outputs = judge.judge(judge_inputs)
    return judge_outputs


def judge_conversations(
    config: JudgeConfig, judge_inputs: list[Conversation]
) -> list[dict[str, Any]]:
    """Judge a list of conversations.

    This function evaluates a list of conversations using the specified Judge.

    The function performs the following steps:

        1. Initializes the Judge with the provided configuration.
        2. Uses the Judge to evaluate each conversation input.
        3. Collects and returns the judged outputs.

    Args:
        config: The configuration for the judge.
        judge_inputs: A list of Conversation objects to be judged.

    Returns:
        List[Dict[str, Any]]: A list of judgement results for each conversation.

        >>> # Example output:
        [
            {'helpful': True, 'safe': False},
            {'helpful': True, 'safe': True},
        ]

    Example:
        >>> config = JudgeConfig(...)
        >>> judge_inputs = [Conversation(...), Conversation(...)]
        >>> judged_outputs = judge_conversations(config, judge_inputs)
        >>> for output in judged_outputs:
        ...     print(output)
    """
    judge = Judge(config)
    judge_outputs = judge.judge(judge_inputs)
    return judge_outputs



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
import torch
import transformers

from oumi.core.constants import LABEL_IGNORE_INDEX
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types import Conversation
from oumi.utils.logging import logger


def tokenize_for_completions_only_training_with_template(
    tokenizer: BaseTokenizer, conversation: Conversation
) -> dict:
    """Tokenize a conversation for completions-only training with a template."""
    batch: transformers.BatchEncoding = tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=True,
        return_dict=True,
        return_assistant_tokens_mask=True,
    )

    data = batch.data

    assistant_tokens_mask = data.pop("assistant_masks")

    data["labels"] = [
        token_id if mask else LABEL_IGNORE_INDEX
        for mask, token_id in zip(assistant_tokens_mask, data["input_ids"])
    ]

    return data


def tokenize_for_completions_only_training_with_prefix(
    tokenizer: BaseTokenizer,
    conversation: Conversation,
    response_template: str,
    instruction_template: str,
    response_token_ids: list[int],
    instruction_token_ids: list[int],
) -> dict:
    """Tokenize a conversation for completions-only training with a prefix."""
    prompt: str = tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=False,
        return_dict=False,
        return_assistant_tokens_mask=False,
    )
    tokenizer_batch: transformers.BatchEncoding = tokenizer(
        prompt, truncation=True, padding=False, return_tensors="pt"
    )

    batch = {k: v[0] for k, v in tokenizer_batch.data.items()}
    batch["labels"] = batch["input_ids"].clone()

    response_token_ids_idxs = []
    human_token_ids_idxs = []

    cond = np.atleast_1d(batch["labels"] == response_token_ids[0])

    for assistant_idx in np.where(cond)[0]:
        # find the indexes of the start of a response.
        if (
            response_token_ids
            == batch["labels"][
                assistant_idx : assistant_idx + len(response_token_ids)
            ].tolist()
        ):
            response_token_ids_idxs.append(assistant_idx + len(response_token_ids))

    if len(response_token_ids_idxs) == 0:
        logger.warning(
            f"Could not find response key `{response_template}` in the "
            f'following instance: {tokenizer.decode(batch["input_ids"])} '
            f"This instance will be ignored in loss calculation. "
            f"Note, if this happens often, consider increasing the `max_seq_length`."
        )
        batch["labels"][:] = LABEL_IGNORE_INDEX

    human_token_ids = instruction_token_ids
    for human_idx in np.where(batch["labels"] == human_token_ids[0])[0]:
        # find the indexes of the start of a human answer.
        if (
            human_token_ids
            == batch["labels"][human_idx : human_idx + len(human_token_ids)].tolist()
        ):
            human_token_ids_idxs.append(human_idx)

    if len(human_token_ids_idxs) == 0:
        logger.warn(
            f"Could not find instruction key `{instruction_template}` in the "
            f'following instance: {tokenizer.decode(batch["input_ids"])} '
            f"This instance will be ignored in loss calculation. "
            f"Note, if this happens often, consider increasing the `max_seq_length`."
        )
        batch["labels"][:] = LABEL_IGNORE_INDEX

    if (
        len(human_token_ids_idxs) > 0
        and len(response_token_ids_idxs) > 0
        and human_token_ids_idxs[0] > response_token_ids_idxs[0]
    ):
        human_token_ids_idxs = [0] + human_token_ids_idxs

    for idx, (start, end) in enumerate(
        zip(human_token_ids_idxs, response_token_ids_idxs)
    ):
        # Make pytorch loss function ignore all non response tokens
        if idx != 0:
            batch["labels"][start:end] = LABEL_IGNORE_INDEX
        else:
            batch["labels"][:end] = LABEL_IGNORE_INDEX

    if len(response_token_ids_idxs) < len(human_token_ids_idxs):
        batch["labels"][human_token_ids_idxs[-1] :] = LABEL_IGNORE_INDEX

    return batch


def _find_pattern_start(
    labels: torch.Tensor, pattern_tokens: list[int]
) -> Optional[int]:
    """Find the starting index of the pattern in the labels."""
    # Get all positions where the first token matches
    potential_starts = np.where(np.atleast_1d(labels == pattern_tokens[0]))[0]

    # Check each position for full template match
    for start_idx in potential_starts:
        sequence = labels[start_idx : start_idx + len(pattern_tokens)].tolist()
        if sequence == pattern_tokens:
            return start_idx

    return None


def tokenizer_for_inference(
    tokenizer: BaseTokenizer, conversation: Conversation
) -> dict:
    """Tokenize a conversation for inference."""
    return tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=True,
        return_dict=True,
    )



from typing import Optional
from unittest.mock import Mock, patch

import pytest
from torch import nn

from oumi.builders.models import (
    _get_model_type,
    _patch_model_for_liger_kernel,
    build_chat_template,
    build_tokenizer,
    is_image_text_llm,
)
from oumi.core.configs import ModelParams
from oumi.utils.logging import logger


@pytest.fixture
def mock_liger_kernel():
    with patch("oumi.builders.models.liger_kernel") as mock:
        yield mock


def create_mock_model(model_type):
    model = Mock(spec=nn.Module)
    model.config = Mock()
    model.config.model_type = model_type
    return model


@pytest.mark.parametrize(
    "model_type",
    [
        "llama",
        "mixtral",
        "mistral",
        "gemma",
    ],
)
def test_patch_model_for_liger_kernel(mock_liger_kernel, model_type):
    model = create_mock_model(model_type)

    _patch_model_for_liger_kernel(model)

    mock_liger_kernel.transformers._apply_liger_kernel.assert_called_once_with(
        model_type
    )


def test_patch_model_for_liger_kernel_no_config(mock_liger_kernel):
    model = Mock(spec=nn.Module)
    with pytest.raises(ValueError, match=f"Could not find model type for: {model}"):
        _patch_model_for_liger_kernel(model)


def test_patch_model_for_liger_kernel_import_error():
    with patch("oumi.builders.models.liger_kernel", None):
        model = create_mock_model("llama")
        with pytest.raises(ImportError, match="Liger Kernel not installed"):
            _patch_model_for_liger_kernel(model)


def test_get_model_type():
    # Test with valid model
    model = create_mock_model("llama")
    assert _get_model_type(model) == "llama"

    # Test with no config
    model = Mock(spec=nn.Module)
    assert _get_model_type(model) is None

    # Test with config but no model_type
    model = Mock(spec=nn.Module)
    model.config = Mock()
    model.config.model_type = None
    assert _get_model_type(model) is None


@pytest.mark.parametrize(
    "template_name",
    ["zephyr"],
)
def test_build_chat_template_existing_templates(template_name):
    template = build_chat_template(template_name)

    assert template is not None
    assert len(template) > 0


def test_build_chat_template_nonexistent_template():
    with pytest.raises(FileNotFoundError) as exc_info:
        build_chat_template("nonexistent_template")

    assert "Chat template file not found" in str(exc_info.value)


def test_build_chat_template_removes_indentation_and_newlines():
    template_content = """
        {{ bos_token }}
        {% for message in messages %}
            {% if message['role'] == 'user' %}
                User: {{ message['content'] }}
            {% elif message['role'] == 'assistant' %}
                Assistant: {{ message['content'] }}
            {% endif %}
            {{ eos_token }}
        {% endfor %}
    """
    expected = (
        "{{ bos_token }}{% for message in messages %}{% if message['role'] == 'user' %}"
        "User: {{ message['content'] }}{% elif message['role'] == 'assistant' %}"
        "Assistant: {{ message['content'] }}{% endif %}{{ eos_token }}{% endfor %}"
    )

    with (
        patch("oumi.builders.models.get_oumi_root_directory"),
        patch("oumi.builders.models.load_file") as mock_load_file,
    ):
        mock_load_file.return_value = template_content

        result = build_chat_template("test_template")

        assert result == expected
        mock_load_file.assert_called_once()


@pytest.mark.parametrize(
    "model_name, trust_remote_code, expected_result",
    [
        ("MlpEncoder", False, False),  # Custom model
        ("CnnClassifier", False, False),  # Custom model
        ("openai-community/gpt2", False, False),
        ("HuggingFaceTB/SmolLM2-135M-Instruct", False, False),
        ("llava-hf/llava-1.5-7b-hf", False, True),
        ("Salesforce/blip2-opt-2.7b", False, True),
        ("microsoft/Phi-3-vision-128k-instruct", True, True),
        # ("HuggingFaceTB/SmolVLM-Instruct", False, True), # requires transformers>=4.46
    ],
)
def test_is_image_text_llm(
    model_name: str, trust_remote_code: bool, expected_result: bool
):
    assert (
        is_image_text_llm(
            ModelParams(model_name=model_name, trust_remote_code=trust_remote_code)
        )
        == expected_result
    )


@pytest.mark.parametrize(
    "model_name, trust_remote_code, template_name, expected_padding_side",
    [
        ("openai-community/gpt2", False, "gpt2", "right"),
        ("HuggingFaceTB/SmolLM2-135M-Instruct", False, None, "right"),
        ("llava-hf/llava-1.5-7b-hf", False, "llava", "left"),
        ("microsoft/Phi-3-vision-128k-instruct", True, "phi3-instruct", "right"),
        ("Qwen/Qwen2-VL-2B-Instruct", True, "qwen2-vl-instruct", "left"),
        # These models require allowlisting:
        # ("meta-llama/Llama-3.2-3B-Instruct", False, None, "right"),
    ],
)
def test_default_chat_template_in_build_tokenizer(
    model_name: str,
    trust_remote_code: bool,
    template_name: Optional[str],
    expected_padding_side: str,
):
    tokenizer = build_tokenizer(
        ModelParams(model_name=model_name, trust_remote_code=trust_remote_code)
    )

    debug_tag = f"template_name: {template_name} model_name: {model_name}"
    if template_name:
        expected = build_chat_template(template_name=template_name)
        if tokenizer.chat_template != expected:
            logger.info(
                f"Tokenizer chat template:\n\n{tokenizer.chat_template}\n\n"
                f"Expected chat template:\n\n{expected}\n\n"
            )
        assert tokenizer.chat_template == expected, debug_tag
    else:
        # Using the model's built-in config.
        assert (
            tokenizer.chat_template is not None
        ), f"Unspecified built-in template: {debug_tag}"
        assert len(tokenizer.chat_template) > 0, f"Empty built-in template: {debug_tag}"

    # Also check padding side here.
    assert hasattr(tokenizer, "padding_side")
    assert tokenizer.padding_side == expected_padding_side



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Generic, Optional, TypeVar

import pydantic

from oumi.core.configs import BaseConfig
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.params.generation_params import GenerationParams
from oumi.core.configs.params.model_params import ModelParams
from oumi.core.configs.params.remote_params import RemoteParams
from oumi.core.types.conversation import Conversation, Message, Role, TemplatedMessage


class JudgeAttributeValueType(str, Enum):
    """Enumeration of possible value types for judge attributes."""

    BOOL = "bool"
    """Boolean value type."""

    CATEGORICAL = "categorical"
    """Categorical value type."""

    LIKERT_5 = "likert-5"
    """Likert scale with 5 points value type."""


T = TypeVar("T", bound=TemplatedMessage)


class JudgeAttribute(pydantic.BaseModel, Generic[T]):
    """Attributes for the judge.

    Example:
        >>> attribute = JudgeAttribute( # doctest: +SKIP
        ...     name="helpful",
        ...     system_prompt="You are an impartial judge.",
        ...     examples=[
        ...         TemplatedMessage(
        ...             role=Role.USER,
        ...             request="What is the capital of France?",
        ...             response="The capital of France is Paris.",
        ...         ),
        ...         TemplatedMessage(
        ...             role=Role.ASSISTANT,
        ...             response="True",
        ...         ),
        ...     ],
        ...     value_type=JudgeAttributeValueType.BOOL,
        ...     limit_examples=5,
        ... )
        >>> print(attribute.name) # doctest: +SKIP
        helpful
    """

    name: str
    """The name of the attribute being judged."""

    system_prompt: str
    """The system prompt for the judge."""

    examples: list[T] = field(default_factory=list)
    """A list of few-shot example inputs and judgements."""

    value_type: JudgeAttributeValueType = JudgeAttributeValueType.BOOL
    """The type of value for the attribute."""

    limit_examples: Optional[int] = 5
    """The maximum number of examples to use.

    This is an optional parameter that limits the number of examples to be used for
    judging the attribute. If not specified, the default is 5.
    """

    @property
    def conversation(self) -> Conversation:
        """Returns the judgement conversation in oumi format.

        This will include the judge system prompt, and any few-shot examples.
        """
        return Conversation(messages=self.messages)

    @property
    def messages(self) -> list[Message]:
        """Returns the messages in oumi format.

        This will include the judge system prompt, and any few-shot examples.
        """
        messages = [Message(content=self.system_prompt, role=Role.SYSTEM)]
        return messages + [e.message for e in self.examples]

    @classmethod
    def load(cls: type, filename: str) -> "JudgeAttribute[T]":
        """Loads the judge attribute config from a file."""
        path = Path(filename)
        if not path.exists():
            raise FileNotFoundError(path)
        return cls.model_validate_json(path.read_text())


@dataclass
class JudgeConfig(BaseConfig):
    """Configuration for the Judge.

    This class holds the configuration for the Judge,
      including the attributes to judge, the model parameters,
      and the text generation parameters.

    Examples:
        >>> attributes = {
        ...     "helpful": JudgeAttribute( # doctest: +SKIP
        ...         name="helpful",
        ...         system_prompt="Is this answer helpful?",
        ...         examples=[
        ...             TemplatedMessage(
        ...                 role=Role.USER,
        ...                 request="What is the capital of France?",
        ...                 response="The capital of France is Paris.",
        ...             ),
        ...             TemplatedMessage(
        ...                 role=Role.ASSISTANT,
        ...                 response="True",
        ...             ),
        ...         ],
        ...     ),
        ...     "honest": JudgeAttribute(
        ...         name="honest",
        ...         system_prompt="Is this answer honest?",
        ...         examples=[]
        ...     )
        ... }
        >>> model_params = ModelParams(model_name="example-model")
        >>> generation_params = GenerationParams(max_new_tokens=100) # doctest: +SKIP
        >>> judge_config = JudgeConfig( # doctest: +SKIP
        ...     attributes=attributes,
        ...     model=model_params,
        ...     generation=generation_params
        ... )
    """

    attributes: dict[str, JudgeAttribute] = field(default_factory=dict)
    """The attributes to judge."""

    model: ModelParams = field(default_factory=ModelParams)
    """Parameters for the model used in inference."""

    generation: GenerationParams = field(default_factory=GenerationParams)
    """Parameters for text generation during inference."""

    engine: InferenceEngineType = field(default=InferenceEngineType.NATIVE)
    """The inference engine to use for generation."""

    remote_params: Optional[RemoteParams] = None
    """Parameters for running inference against a remote API."""



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from oumi.core.configs import (
    GenerationParams,
    JudgeConfig,
    ModelParams,
    RemoteParams,
)
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.judge_config import JudgeAttribute
from oumi.core.registry import register_judge
from oumi.judges.oumi_judge import OumiJudgeInput, OumiJudgeOutput
from oumi.utils.io_utils import get_oumi_root_directory


@register_judge("oumi/v1_xml_claude_sonnet")
def oumi_v1_xml_claude_sonnet_judge() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML Anthropic judge.

    This function creates and returns a JudgeConfig object for the Oumi V1 Judge, which
    uses Claude Sonnet as a judge, with inputs and outputs in XML format.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML Anthropic judge.

    Note:
        This judge uses the Anthropic API, so the ANTHROPIC_API_KEY environment
        variable must be set with a valid API key.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }

    config = JudgeConfig(
        attributes=attributes,
        model=ModelParams(
            model_name="claude-3-5-sonnet-20240620",
        ),
        generation=GenerationParams(
            max_new_tokens=1024,
        ),
        remote_params=RemoteParams(
            api_url="https://api.anthropic.com/v1/messages",
            api_key_env_varname="ANTHROPIC_API_KEY",
            max_retries=3,
        ),
        engine=InferenceEngineType.ANTHROPIC,
    )
    return config


@register_judge("oumi/v1_xml_local")
def oumi_v1_xml_local_judge() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML local judge.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML local judge.

    Note:
        This judge uses a local GGUF model file for inference.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }
    config = JudgeConfig(
        attributes=attributes,
        model=ModelParams(model_name="Qwen/Qwen2-0.5B-Instruct-GGUF"),
        engine=InferenceEngineType.LLAMACPP,
        generation=GenerationParams(max_new_tokens=1024, temperature=0.0),
    )
    return config


@register_judge("oumi/v1_xml_gpt4o")
def oumi_v1_xml_gpt4o_judge() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML GPT-4 judge.

    This function creates and returns a JudgeConfig object for the Oumi V1 Judge, which
    uses GPT-4 as a judge, with inputs and outputs in XML format.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML GPT-4 judge.

    Note:
        This judge uses the OpenAI API, so the OPENAI_API_KEY environment
        variable must be set with a valid API key.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }

    config = JudgeConfig(
        attributes=attributes,
        model=ModelParams(model_name="gpt-4o-2024-08-06"),
        engine=InferenceEngineType.REMOTE,
        generation=GenerationParams(
            max_new_tokens=1024,
            temperature=0.0,
        ),
        remote_params=RemoteParams(
            api_url="https://api.openai.com/v1/chat/completions",
            api_key_env_varname="OPENAI_API_KEY",
            max_retries=3,
        ),
    )
    return config


@register_judge("oumi/v1_xml_deepseek_r1_hosted_by_deepseek")
def oumi_v1_xml_deepseek_r1_judge_hosted_by_deepseek() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML DeepSeek R1 judge.

    This function creates and returns a JudgeConfig object for the Oumi V1 Judge, which
    uses DeepSeek R1 as a judge, with inputs and outputs in XML format.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML DeepSeek R1 judge.

    Note:
        This judge uses the DeepSeek API, so the DEEPSEEK_API_KEY environment
        variable must be set with a valid API key.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }

    config = JudgeConfig(
        attributes=attributes,
        # refer to model name https://api-docs.deepseek.com/quick_start/pricing
        model=ModelParams(model_name="deepseek-reasoner"),
        engine=InferenceEngineType.DEEPSEEK,
        generation=GenerationParams(
            max_new_tokens=1024,
            temperature=0.0,
        ),
        remote_params=RemoteParams(
            max_retries=3,
        ),
    )
    return config


@register_judge("oumi/v1_xml_deepseek_r1_hosted_by_sambanova")
def oumi_v1_xml_deepseek_r1_judge_hosted_by_sambanova() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML DeepSeek R1 judge.

    This function creates and returns a JudgeConfig object for the Oumi V1 Judge, which
    uses DeepSeek R1 as a judge, with inputs and outputs in XML format.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML DeepSeek R1 judge.

    Note:
        This judge uses the SambaNova Cloud API, so the SAMBANOVA_API_KEY environment
        variable must be set with a valid API key.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }

    config = JudgeConfig(
        attributes=attributes,
        # refer to model name https://api-docs.deepseek.com/quick_start/pricing
        model=ModelParams(model_name="DeepSeek-R1"),
        engine=InferenceEngineType.SAMBANOVA,
        generation=GenerationParams(
            max_new_tokens=1024,
            temperature=0.0,
        ),
        remote_params=RemoteParams(
            max_retries=3, connection_timeout=10, politeness_policy=2.0
        ),
    )
    return config


@register_judge("oumi/v1_xml_deepseek_r1_hosted_by_together")
def oumi_v1_xml_deepseek_r1_judge_hosted_by_together() -> JudgeConfig:
    """Returns a JudgeConfig for the Oumi v1 XML DeepSeek R1 judge.

    This function creates and returns a JudgeConfig object for the Oumi V1 Judge, which
    uses DeepSeek R1 as a judge, with inputs and outputs in XML format.

    Returns:
        JudgeConfig: A configuration object for the Oumi v1 XML DeepSeek R1 judge.

    Note:
        This judge uses the Together Cloud API, so the TOGETHER_API_KEY environment
        variable must be set with a valid API key.
    """
    judges_directory = get_oumi_root_directory() / "judges" / "oumi_v1"

    attribute_names = ["helpful", "honest", "safe"]

    attributes = {
        attribute: JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
            str(judges_directory / f"{attribute}.json")
        )
        for attribute in attribute_names
    }

    config = JudgeConfig(
        attributes=attributes,
        model=ModelParams(model_name="deepseek-ai/DeepSeek-R1"),
        engine=InferenceEngineType.TOGETHER,
        generation=GenerationParams(
            max_new_tokens=1024,
            temperature=0.0,
        ),
        remote_params=RemoteParams(
            max_retries=3,
            # Experience number. Based on curl test, it takes about 48 seconds to
            # complete a LLM as judge task
            connection_timeout=60,
            # Note: Due to high demand, DeepSeek R1 has a model specific rate limit of \
            # 6 RPM for Free Users, and 12 RPM for Build Tier 1 users.
            # https://docs.together.ai/docs/rate-limits#rate-limit-tiers
            politeness_policy=11.0,
        ),
    )
    return config


@register_judge("oumi/v1_xml_unit_test")
def unit_test_judge():
    """Tiny judge for unit testing.

    Do not use this judge for anything serious as it returns random results.
    """
    attribute_path = (
        get_oumi_root_directory() / "judges" / "test_judge" / "helpful.json"
    )

    attribute = JudgeAttribute[Union[OumiJudgeInput, OumiJudgeOutput]].load(
        str(attribute_path)
    )

    config = JudgeConfig(
        attributes={"helpful": attribute},
        engine=InferenceEngineType.NATIVE,
        model=ModelParams(
            model_name="gpt2",
            tokenizer_pad_token="</s>",
            chat_template="gpt2",
        ),
        generation=GenerationParams(
            max_new_tokens=128,
            temperature=0.0,
        ),
    )

    return config



from inspect import signature
from typing import Callable, get_type_hints
from unittest.mock import Mock, patch

import pytest
from typer.testing import CliRunner

from oumi.cli.distributed_run import accelerate, torchrun
from oumi.cli.env import env
from oumi.cli.evaluate import evaluate
from oumi.cli.fetch import fetch
from oumi.cli.infer import infer
from oumi.cli.judge import conversations, dataset
from oumi.cli.launch import cancel, down, status, stop, up, which
from oumi.cli.launch import run as launcher_run
from oumi.cli.main import get_app
from oumi.cli.train import train

runner = CliRunner()


def _copy_command(mock: Mock, command: Callable):
    mock.__name__ = command.__name__
    mock.__annotations__ = get_type_hints(command)
    mock.__signature__ = signature(command)
    mock.__bool__ = lambda _: True


#
# Fixtures
#
@pytest.fixture
def mock_train():
    with patch("oumi.cli.main.train") as m_train:
        _copy_command(m_train, train)
        yield m_train


@pytest.fixture
def mock_eval():
    with patch("oumi.cli.main.evaluate") as m_eval:
        _copy_command(m_eval, evaluate)
        yield m_eval


@pytest.fixture
def mock_infer():
    with patch("oumi.cli.main.infer") as m_infer:
        _copy_command(m_infer, infer)
        yield m_infer


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.main.fetch") as m_fetch:
        _copy_command(m_fetch, fetch)
        yield m_fetch


@pytest.fixture
def mock_down():
    with patch("oumi.cli.main.down") as m_down:
        _copy_command(m_down, down)
        yield m_down


@pytest.fixture
def mock_stop():
    with patch("oumi.cli.main.stop") as m_stop:
        _copy_command(m_stop, stop)
        yield m_stop


@pytest.fixture
def mock_launcher_run():
    with patch("oumi.cli.main.launcher_run") as m_launcher_run:
        _copy_command(m_launcher_run, launcher_run)
        yield m_launcher_run


@pytest.fixture
def mock_status():
    with patch("oumi.cli.main.status") as m_status:
        _copy_command(m_status, status)
        yield m_status


@pytest.fixture
def mock_cancel():
    with patch("oumi.cli.main.cancel") as m_cancel:
        _copy_command(m_cancel, cancel)
        yield m_cancel


@pytest.fixture
def mock_up():
    with patch("oumi.cli.main.up") as m_up:
        _copy_command(m_up, up)
        yield m_up


@pytest.fixture
def mock_which():
    with patch("oumi.cli.main.which") as m_which:
        _copy_command(m_which, which)
        yield m_which


@pytest.fixture
def mock_judge_dataset():
    with patch("oumi.cli.main.dataset") as m_dataset:
        _copy_command(m_dataset, dataset)
        yield m_dataset


@pytest.fixture
def mock_judge_conversations():
    with patch("oumi.cli.main.conversations") as m_conversations:
        _copy_command(m_conversations, conversations)
        yield m_conversations


@pytest.fixture
def mock_distributed_torchrun():
    with patch("oumi.cli.main.torchrun") as m_torchrun:
        _copy_command(m_torchrun, torchrun)
        yield m_torchrun


@pytest.fixture
def mock_distributed_accelerate():
    with patch("oumi.cli.main.accelerate") as m_accelerate:
        _copy_command(m_accelerate, accelerate)
        yield m_accelerate


@pytest.fixture
def mock_env():
    with patch("oumi.cli.main.env") as m_env:
        _copy_command(m_env, env)
        yield m_env


def test_main_train_registered(mock_train):
    _ = runner.invoke(
        get_app(), ["train", "--config", "some/path", "--allow_extra" "args"]
    )
    mock_train.assert_called_once()


def test_main_infer_registered(mock_infer):
    _ = runner.invoke(
        get_app(), ["infer", "--config", "some/path", "--allow_extra" "args"]
    )
    mock_infer.assert_called_once()


def test_main_fetch_registered(mock_fetch):
    _ = runner.invoke(get_app(), ["fetch", "some/path", "--output-dir", "output/path"])
    mock_fetch.assert_called_once()


def test_main_eval_registered(mock_eval):
    _ = runner.invoke(
        get_app(), ["eval", "--config", "some/path", "--allow_extra" "args"]
    )
    mock_eval.assert_called_once()


def test_main_evaluate_registered(mock_eval):
    _ = runner.invoke(
        get_app(), ["evaluate", "--config", "some/path", "--allow_extra" "args"]
    )
    mock_eval.assert_called_once()


def test_main_launch_registered():
    result = runner.invoke(get_app(), ["launch", "--help"])
    for cmd in ["down", "stop", "run", "status", "cancel", "up", "which"]:
        assert cmd in result.output


def test_main_down_registered(mock_down):
    _ = runner.invoke(
        get_app(), ["launch", "down", "--cluster", "cluster", "--cloud", "gcp"]
    )
    mock_down.assert_called_once()


def test_main_stop_registered(mock_stop):
    _ = runner.invoke(
        get_app(), ["launch", "stop", "--cluster", "cluster", "--cloud", "gcp"]
    )
    mock_stop.assert_called_once()


def test_main_run_registered(mock_launcher_run):
    _ = runner.invoke(
        get_app(),
        ["launch", "run", "--config", "some_path", "--cluster", "clust", "--detach"],
    )
    mock_launcher_run.assert_called_once()


def test_main_status_registered(mock_status):
    _ = runner.invoke(
        get_app(),
        [
            "launch",
            "status",
            "--cloud",
            "gcp",
            "--cluster",
            "cluster",
            "--id",
            "foobar",
        ],
    )
    mock_status.assert_called_once()


def test_main_cancel_registered(mock_cancel):
    _ = runner.invoke(
        get_app(),
        [
            "launch",
            "cancel",
            "--cloud",
            "gcp",
            "--cluster",
            "cluster",
            "--id",
            "foobar",
        ],
    )
    mock_cancel.assert_called_once()


def test_main_up_registered(mock_up):
    _ = runner.invoke(
        get_app(),
        ["launch", "up", "--config", "some_path", "--cluster", "clust", "--detach"],
    )
    mock_up.assert_called_once()


def test_main_which_registered(mock_which):
    _ = runner.invoke(get_app(), ["launch", "which"])
    mock_which.assert_called_once()


def test_main_judge_dataset_registered(mock_judge_dataset):
    _ = runner.invoke(
        get_app(),
        [
            "judge",
            "dataset",
            "--config",
            "some_config",
            "--dataset-name",
            "some_dataset",
        ],
    )
    mock_judge_dataset.assert_called_once()


def test_main_env_registered(mock_env):
    _ = runner.invoke(get_app(), ["env"])
    mock_env.assert_called_once()


def test_main_judge_conversations_registered(mock_judge_conversations):
    _ = runner.invoke(
        get_app(),
        [
            "judge",
            "conversations",
            "--config",
            "some_config",
            "--input-file",
            "some_file.jsonl",
        ],
    )
    mock_judge_conversations.assert_called_once()


def test_main_distributed_registered():
    result = runner.invoke(get_app(), ["distributed", "--help"])
    for cmd in ["accelerate", "torchrun"]:
        assert cmd in result.output


def test_main_distributed_torchrun_registered(mock_distributed_torchrun):
    _ = runner.invoke(
        get_app(),
        [
            "distributed",
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--config",
            "some_config",
        ],
    )
    mock_distributed_torchrun.assert_called_once()


def test_main_distributed_accelerate_registered(mock_distributed_accelerate):
    _ = runner.invoke(
        get_app(),
        [
            "distributed",
            "accelerate",
            "-m",
            "oumi",
            "train",
            "--config",
            "some_config",
        ],
    )
    mock_distributed_accelerate.assert_called_once()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional

import numpy as np

from oumi.core.configs.params.base_params import BaseParams


@dataclass
class RemoteParams(BaseParams):
    """Parameters for running inference against a remote API."""

    api_url: Optional[str] = None
    """URL of the API endpoint to use for inference."""

    api_key: Optional[str] = None
    """API key to use for authentication."""

    api_key_env_varname: Optional[str] = None
    """Name of the environment variable containing the API key for authentication."""

    max_retries: int = 3
    """Maximum number of retries to attempt when calling an API."""

    connection_timeout: float = 20.0
    """Timeout in seconds for a request to an API."""

    num_workers: int = 1
    """Number of workers to use for parallel inference."""

    politeness_policy: float = 0.0
    """Politeness policy to use when calling an API.

    If greater than zero, this is the amount of time in seconds a worker will sleep
    before making a subsequent request.
    """

    batch_completion_window: Optional[str] = "24h"
    """Time window for batch completion. Currently only '24h' is supported.

    Only used for batch inference.
    """

    def __post_init__(self):
        """Validate the remote parameters."""
        if self.num_workers < 1:
            raise ValueError(
                "Number of num_workers must be greater than or equal to 1."
            )
        if self.politeness_policy < 0:
            raise ValueError("Politeness policy must be greater than or equal to 0.")
        if self.connection_timeout < 0:
            raise ValueError("Connection timeout must be greater than or equal to 0.")
        if not np.isfinite(self.politeness_policy):
            raise ValueError("Politeness policy must be finite.")
        if self.max_retries < 0:
            raise ValueError("Max retries must be greater than or equal to 0.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import logging
import re
from collections.abc import Iterator
from io import StringIO
from pathlib import Path
from typing import Any, Optional, TypeVar, Union, cast

from omegaconf import OmegaConf

from oumi.core.configs.params.base_params import BaseParams

T = TypeVar("T", bound="BaseConfig")

_CLI_IGNORED_PREFIXES = ["--local-rank"]


def _filter_ignored_args(arg_list: list[str]) -> list[str]:
    """Filters out ignored CLI arguments."""
    return [
        arg
        for arg in arg_list
        if not any(arg.startswith(prefix) for prefix in _CLI_IGNORED_PREFIXES)
    ]


def _read_config_without_interpolation(config_path: str) -> str:
    """Reads a configuration file without interpolating variables.

    Args:
        config_path: The path to the configuration file.

    Returns:
        str: The stringified configuration.
    """
    with open(config_path) as f:
        stringified_config = f.read()
        pattern = r"(?<!\\)\$\{"  # Matches "${" but not "\${"
        stringified_config = re.sub(pattern, "\\${", stringified_config)
    return stringified_config


@dataclasses.dataclass
class BaseConfig:
    def to_yaml(self, config_path: Union[str, Path, StringIO]) -> None:
        """Saves the configuration to a YAML file."""
        OmegaConf.save(config=self, f=config_path)

    @classmethod
    def from_yaml(
        cls: type[T], config_path: Union[str, Path], ignore_interpolation=True
    ) -> T:
        """Loads a configuration from a YAML file.

        Args:
            config_path: The path to the YAML file.
            ignore_interpolation: If True, then any interpolation variables in the
                configuration file will be escaped.

        Returns:
            BaseConfig: The merged configuration object.
        """
        schema = OmegaConf.structured(cls)
        if ignore_interpolation:
            stringified_config = _read_config_without_interpolation(str(config_path))
            file_config = OmegaConf.create(stringified_config)
        else:
            file_config = OmegaConf.load(config_path)
        config = OmegaConf.to_object(OmegaConf.merge(schema, file_config))
        if not isinstance(config, cls):
            raise TypeError(f"config is not {cls}")
        return cast(T, config)

    @classmethod
    def from_str(cls: type[T], config_str: str) -> T:
        """Loads a configuration from a YAML string.

        Args:
            config_str: The YAML string.

        Returns:
            BaseConfig: The configuration object.
        """
        schema = OmegaConf.structured(cls)
        file_config = OmegaConf.create(config_str)
        config = OmegaConf.to_object(OmegaConf.merge(schema, file_config))
        if not isinstance(config, cls):
            raise TypeError(f"config is not {cls}")
        return cast(T, config)

    @classmethod
    def from_yaml_and_arg_list(
        cls: type[T],
        config_path: Optional[str],
        arg_list: list[str],
        logger: Optional[logging.Logger] = None,
        ignore_interpolation=True,
    ) -> T:
        """Loads a configuration from various sources.

        If both YAML and arguments list are provided, then
        parameters specified in `arg_list` have higher precedence.

        Args:
            config_path: The path to the YAML file.
            arg_list: Command line arguments list.
            logger: (optional) Logger.
            ignore_interpolation: If True, then any interpolation variables in the
                configuration file will be escaped.

        Returns:
            BaseConfig: The merged configuration object.
        """
        # Start with an empty typed config. This forces OmegaConf to validate
        # that all other configs are of this structured type as well.
        all_configs = [OmegaConf.structured(cls)]

        # Override with configuration file if provided.
        if config_path is not None:
            if ignore_interpolation:
                stringified_config = _read_config_without_interpolation(config_path)
                all_configs.append(OmegaConf.create(stringified_config))
            else:
                all_configs.append(cls.from_yaml(config_path))

        # Merge base config and config from yaml.
        try:
            # Merge and validate configs
            config = OmegaConf.merge(*all_configs)
        except Exception:
            if logger:
                configs_str = "\n\n".join([f"{config}" for config in all_configs])
                logger.exception(
                    f"Failed to merge {len(all_configs)} Omega configs:\n{configs_str}"
                )
            raise

        # Override config with CLI arguments, in order. The arguments, aka flag names,
        # are dot-separated arguments, ex. `model.model_name`. This also supports
        # arguments indexing into lists, ex. `tasks[0].num_samples` or
        # `tasks.0.num_samples`. This is because the config is already populated and
        # typed, so the indexing is properly interpreted as a list index as opposed to
        # a dictionary key.
        try:
            # Filter out CLI arguments that should be ignored.
            arg_list = _filter_ignored_args(arg_list)
            # Override with CLI arguments.
            config.merge_with_dotlist(arg_list)
        except Exception:
            if logger:
                logger.exception(
                    f"Failed to merge arglist {arg_list} with Omega config:\n{config}"
                )
            raise

        config = OmegaConf.to_object(config)
        if not isinstance(config, cls):
            raise TypeError(f"config {type(config)} is not {type(cls)}")

        return cast(T, config)

    def finalize_and_validate(self) -> None:
        """Finalizes and validates the top level params objects."""
        for _, attr_value in self:
            if isinstance(attr_value, BaseParams):
                attr_value.finalize_and_validate()

        self.__finalize_and_validate__()

    def __finalize_and_validate__(self) -> None:
        """Finalizes and validates the parameters of this object.

        This method can be overridden by subclasses to implement custom
        validation logic.

        In case of validation errors, this method should raise a `ValueError`
        or other appropriate exception.
        """

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        """Returns an iterator over field names and values.

        Note: for an attribute to be a field, it must be declared in the
        dataclass definition and have a type annotation.
        """
        for param in dataclasses.fields(self):
            yield param.name, getattr(self, param.name)



import unittest

import pytest

from oumi.core.types.conversation import Conversation, Message, Role
from oumi.judges.judge_court import (
    oumi_v1_xml_deepseek_r1_judge_hosted_by_deepseek,
    oumi_v1_xml_deepseek_r1_judge_hosted_by_sambanova,
    oumi_v1_xml_deepseek_r1_judge_hosted_by_together,
)
from oumi.judges.oumi_judge import OumiXmlJudge as OumiJudge


class TestDeepSeekR1Judge(unittest.TestCase):
    @pytest.mark.skip("Skipping this test due to unable to add funds")
    def test_deepseek_r1_judge_config_hosted_by_deepseek(self):
        conversations = [
            Conversation(
                messages=[
                    Message(
                        role=Role.USER, content="What is the sum of 1 and 1 in binary?"
                    ),
                    Message(role=Role.ASSISTANT, content="The sum is 11 in binary."),
                ]
            ),
            Conversation(
                messages=[
                    Message(role=Role.USER, content="What's the capital of France?"),
                    Message(role=Role.ASSISTANT, content="French people love Paris!"),
                ]
            ),
        ]

        # Get the judge configuration
        config = oumi_v1_xml_deepseek_r1_judge_hosted_by_deepseek()
        judge = OumiJudge(config)
        judge_output = judge.judge(conversations)

        # Print the results
        print(judge_output)

    @pytest.mark.skip("Skipping this test due to unable to R1 API on waiting list")
    def test_deepseek_r1_judge_config_hosted_by_sambanova(self):
        conversations = [
            Conversation(
                messages=[
                    Message(
                        role=Role.USER, content="What is the sum of 1 and 1 in binary?"
                    ),
                    Message(role=Role.ASSISTANT, content="The sum is 11 in binary."),
                ]
            ),
            Conversation(
                messages=[
                    Message(role=Role.USER, content="What's the capital of France?"),
                    Message(role=Role.ASSISTANT, content="French people love Paris!"),
                ]
            ),
        ]

        # Get the judge configuration
        config = oumi_v1_xml_deepseek_r1_judge_hosted_by_sambanova()
        judge = OumiJudge(config)
        judge_output = judge.judge(conversations)

        # Print the results
        print(judge_output)

    def test_deepseek_r1_judge_config_hosted_by_together(self):
        conversations = [
            Conversation(
                messages=[
                    Message(
                        role=Role.USER, content="What is the sum of 1 and 1 in binary?"
                    ),
                    Message(role=Role.ASSISTANT, content="The sum is 11 in binary."),
                ]
            ),
            Conversation(
                messages=[
                    Message(role=Role.USER, content="What's the capital of France?"),
                    Message(role=Role.ASSISTANT, content="French people love Paris!"),
                ]
            ),
        ]

        # Get the judge configuration
        config = oumi_v1_xml_deepseek_r1_judge_hosted_by_together()
        judge = OumiJudge(config)
        judge_output = judge.judge(conversations)

        # Print the results
        print(judge_output)


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import re
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional

from oumi.core.launcher import JobStatus
from oumi.utils.logging import logger

_CTRL_PATH = "-S ~/.ssh/control-%h-%p-%r"


class _SlurmAuthException(Exception):
    pass


def _check_connection(user: str, slurm_host: str) -> None:
    """Checks if the connection is still open."""
    ssh_cmd = f"ssh {_CTRL_PATH} -O check {user}@{slurm_host}"
    error_msg = ""
    try:
        child = subprocess.run(
            ssh_cmd,
            shell=True,
            capture_output=True,
            timeout=10,
        )
        error_msg = child.stderr.decode("utf-8")
    except subprocess.TimeoutExpired:
        raise _SlurmAuthException("Timeout while checking connection.")
    if child.returncode == 0:
        return
    if error_msg:
        logger.error(f"Error checking connection: {error_msg}")
        error_msg = f" Error: {error_msg}"
    raise _SlurmAuthException("Connection to Slurm host is closed." + error_msg)


def _is_job_done(job_state: str) -> bool:
    """Determines if a job is done based on its state.

    See https://slurm.schedmd.com/job_state_codes.html for more details.

    Args:
        job_state: The state of the job.

    Returns:
        True if the job is done, False otherwise.
    """
    terminal_states = {
        "BOOT_FAIL",
        "CANCELLED",
        "COMPLETED",
        "DEADLINE",
        "FAILED",
        "LAUNCH_FAILED",
        "NODE_FAIL",
        "OUT_OF_MEMORY",
        "PREEMPTED",
        "TIMEOUT",
        "SUSPENDED",
        "STOPPED",
    }
    return job_state in terminal_states


def _split_status_line(
    line: str, column_lengths: list[int], cluster_name: str, metadata: str
) -> JobStatus:
    """Splits a status line into a JobStatus object.

    The expected order of job fields is:
    0. Job ID
    1. Job Name
    2. User
    3. Job State
    4. Job State Reason

    Sample status report:
    ID      NAME   USER     STATE REASON
    ----- ------ ------ --------- ------
    1     my_job   user COMPLETED    0:0
    2       job2          RUNNING

    Args:
        line: The line to split.
        column_lengths: The lengths in chars of each column in the line.
        cluster_name: The name of the cluster the job is running on.
        metadata: Additional metadata to attach to the job status.

    Returns:
        A JobStatus object.
    """
    if len(column_lengths) != 5:
        raise ValueError(
            f"Expected 5 fields, but found {len(column_lengths)}."
            f" Invalid line: {line}."
        )
    fields = []
    # Note: We can't use a simple split() here because empty fields are allowed.
    for i in range(len(column_lengths)):
        start = sum(column_lengths[:i]) + i
        end = start + column_lengths[i]
        fields.append(line[start:end].strip())
    return JobStatus(
        id=fields[0],
        name=fields[1],
        # JobState can have additional information. The primary state is the first word.
        status=fields[3].split(" ")[0],
        cluster=cluster_name,
        metadata=metadata,
        done=_is_job_done(fields[3]),
    )


def _compute_duration_debug_str(start_time: float) -> str:
    duration_sec = time.perf_counter() - start_time
    return f"Duration: {duration_sec:.2f} sec"


@dataclass
class SlurmResponse:
    """A response from Slurm."""

    stdout: str
    stderr: str
    exit_code: int


def retry_auth(user_function: Callable) -> Callable:
    """Decorator to ensure auth is fresh before calling a function."""

    @functools.wraps(user_function)
    def wrapper(self, *args, **kwargs):
        self._refresh_creds()
        return user_function(self, *args, **kwargs)

    return wrapper


class SlurmClient:
    """A client for communicating with a Slurm cluster."""

    def __init__(self, user: str, slurm_host: str, cluster_name: str):
        """Initializes a new instance of the SlurmClient class.

        Args:
            user: The user to act as.
            slurm_host: The Slurm Head Node to connect to.
            cluster_name: The name of the cluster this client communicates with.
        """
        self._user = user
        self._slurm_host = slurm_host
        self._cluster_name = cluster_name
        self._refresh_creds()

    def _refresh_creds(self):
        """Refreshes the credentials for the client."""
        try:
            _check_connection(self._user, self._slurm_host)
            # We have fresh credentials, so we return early.
            return
        except _SlurmAuthException:
            logger.warning("No connection found. Establishing a new SSH tunnel...")
        ssh_cmd = (
            f'ssh -f -N -M {_CTRL_PATH} -o "ControlPersist 4h" '
            f"{self._user}@{self._slurm_host}"
        )
        child = subprocess.run(
            ssh_cmd,
            shell=True,
            capture_output=True,
            timeout=180,  # time in seconds
        )
        if child.returncode != 0:
            output = child.stderr.decode("utf-8")
            logger.error(f"Credential error: {output}")
            raise RuntimeError(
                "Failed to refresh Slurm credentials "
                f"for {self._user}@{self._slurm_host}."
            )
        return SlurmResponse(
            stdout=child.stdout.decode("utf-8"),
            stderr=child.stderr.decode("utf-8"),
            exit_code=child.returncode,
        )

    @staticmethod
    def get_active_users(slurm_host: str) -> list[str]:
        """Gets the list of users with an open SSH tunnel to a Slurm cluster.

        Returns:
            A list of users.
        """
        # List all active users with an open SSH tunnel to the Slurm head node.
        command = f"ls ~/.ssh/ | egrep 'control-{slurm_host}-.*-.*'"
        result = subprocess.run(command, shell=True, capture_output=True)
        if result.returncode != 0:
            return []
        # Sample Pattern:
        # control-HOSTNAME-22-taenin
        ssh_tunnel_pattern = rf"control-{slurm_host}-[^-]*-(.*)"
        lines = result.stdout.decode("utf-8").strip().split("\n")
        users = set()
        for line in lines:
            match = re.match(ssh_tunnel_pattern, line.strip())
            if match:
                users.add(match.group(1))
        return list(users)

    def _parse_job_id(self, sbatch_output: str) -> str:
        """Parses the job ID from the result of sbatch.

        From the Slurm MAN page:
        Outputs only the job id number and the cluster name if present.
        The values are separated by a semicolon. Errors will still be displayed.

        Args:
            sbatch_output: The result of sbatch

        Returns:
            The job ID.
        """
        split_job = sbatch_output.strip().split(";")
        if len(split_job) > 2:
            raise ValueError(f"Unexpected output from sbatch: {sbatch_output}")
        return split_job[0]

    @retry_auth
    def run_commands(self, commands: list[str]) -> SlurmResponse:
        """Runs the provided commands in a single SSH command.

        Args:
            commands: The commands to run.
        """
        ssh_cmd = f"ssh {_CTRL_PATH} {self._user}@{self._slurm_host} " " << 'EOF'"
        eof_suffix = "EOF"
        new_cmd = "\n".join([ssh_cmd, *commands, eof_suffix])
        start_time: float = time.perf_counter()
        try:
            logger.debug(f"Running commands:\n{new_cmd}")
            child = subprocess.run(
                new_cmd,
                shell=True,
                capture_output=True,
                timeout=180,  # time in seconds
            )
            duration_str = _compute_duration_debug_str(start_time)
            if child.returncode == 0:
                logger.debug(f"Commands successfully finished! {duration_str}")
            else:
                logger.error(
                    f"Commands failed with code: {child.returncode}! {duration_str}"
                )
            return SlurmResponse(
                stdout=child.stdout.decode("utf-8"),
                stderr=child.stderr.decode("utf-8"),
                exit_code=child.returncode,
            )
        except subprocess.TimeoutExpired:
            duration_str = _compute_duration_debug_str(start_time)
            logger.exception(f"Commands timed out ({duration_str})! {new_cmd}")
            return SlurmResponse(
                stdout="",
                stderr=f"Timeout while running command: {new_cmd}",
                exit_code=1,
            )
        except Exception:
            duration_str = _compute_duration_debug_str(start_time)
            logger.exception(f"Command failed ({duration_str})! {new_cmd}")
            raise

    def submit_job(
        self,
        job_path: str,
        working_dir: str,
        node_count: int,
        name: Optional[str],
    ) -> str:
        """Submits the specified job script to Slurm.

        Args:
            job_path: The path to the job script to submit.
            working_dir: The working directory to submit the job from.
            node_count: The number of nodes to use for the job.
            name: The name of the job (optional).

        Returns:
            The ID of the submitted job.
        """
        optional_name_args = ""
        if name:
            optional_name_args = f"--job-name={name}"
        sbatch_cmd = (
            f"sbatch --nodes={node_count}"
            f" {optional_name_args} --parsable {job_path}"
        )
        result = self.run_commands([f"cd {working_dir}", sbatch_cmd])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to submit job. stderr: {result.stderr}")
        return self._parse_job_id(result.stdout.strip())

    def list_jobs(self) -> list[JobStatus]:
        """Lists all jobs for the current user.

        Returns:
            A list of JobStatus.
        """
        response_format = "JobId%-30,JobName%30,User%30,State%30,Reason%30"
        # Forcibly list all jobs since Jan 1, 2025.
        # Otherwise completed jobs older than ~24 hours may not be listed.
        command = (
            f"sacct --user={self._user} --format='{response_format}' -X "
            "--starttime 2025-01-01"
        )
        result = self.run_commands([command])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to list jobs. stderr: {result.stderr}")
        # Parse STDOUT to retrieve job statuses.
        lines = result.stdout.strip().split("\n")
        jobs = []
        if len(lines) < 2:
            return jobs
        # Look for a line starting in JobID followed by a line starting with "--".
        start_idx = -1
        for idx in range(len(lines) - 1):
            if lines[idx].startswith("JobID") and lines[idx + 1].startswith("--"):
                start_idx = idx
                break
        if start_idx == -1:
            raise RuntimeError(
                f"Failed to parse job list. Unexpected format: {result.stdout}"
            )
        lines = lines[start_idx:]
        # The first two lines are metadata headers.
        # The top line is composed of column titles.
        # The second line is composed of ---- characters, each the length of a column.
        metadata_headers = lines[:2]
        column_lengths = [len(col) for col in lines[1].strip().split(" ")]
        job_lines = lines[2:]
        for line in job_lines:
            job_metadata = "\n".join(metadata_headers + [line])
            status = _split_status_line(
                line, column_lengths, self._cluster_name, job_metadata
            )
            jobs.append(status)
        return jobs

    def get_job(self, job_id: str) -> Optional[JobStatus]:
        """Gets the specified job's status.

        Args:
            job_id: The ID of the job to get.

        Returns:
            The job status if found, None otherwise.
        """
        job_list = self.list_jobs()
        for job in job_list:
            if job.id == job_id:
                return job
        return None

    def cancel(self, job_id) -> Optional[JobStatus]:
        """Cancels the specified job.

        Args:
            job_id: The ID of the job to cancel.

        Returns:
            The job status if found, None otherwise.
        """
        command = f"scancel {job_id}"
        result = self.run_commands([command])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to cancel job. stderr: {result.stderr}")
        return self.get_job(job_id)

    @retry_auth
    def put_recursive(self, source: str, destination: str) -> None:
        """Puts the specified file/directory to the remote path using rsync.

        Args:
            source: The local file/directory to write.
            destination: The remote path to write the file/directory to.
        """
        if Path(source).is_dir():
            self.run_commands([f"mkdir -p {destination}"])
        tests_dir = Path(source) / "tests"
        git_ignore = Path(source) / ".gitignore"
        rsync_cmd_list = [f'rsync -e "ssh {_CTRL_PATH}" -avz --delete ']
        if git_ignore.is_file():
            rsync_cmd_list.append(f"--exclude-from {str(git_ignore)} ")
        if tests_dir.is_dir():
            rsync_cmd_list.append(f"--exclude {str(tests_dir)} ")
        rsync_cmd_list.append(f"{source} ")
        rsync_cmd_list.append(f"{self._user}@{self._slurm_host}:{destination}")
        rsync_cmd = "".join(rsync_cmd_list)
        logger.info(f"Running rsync command: {rsync_cmd} ...")
        try:
            child = subprocess.run(
                rsync_cmd,
                shell=True,
                capture_output=True,
                timeout=300,
            )
            logger.info(f"Rsync command completed with exit code: {child.returncode}")
            if child.returncode != 0:
                parsed_stderr = child.stderr.decode("utf-8") if child.stderr else ""
                raise RuntimeError(f"Rsync failed. stderr: {parsed_stderr}")
        except subprocess.TimeoutExpired:
            raise RuntimeError("Timeout while running rsync command.")

    def put(self, file_contents: str, destination: str) -> None:
        """Puts the specified file contents to the remote path.

        Args:
            file_contents: The contents of the file to write.
            destination: The remote path to write the file to.
        """
        destination_path = Path(destination)
        parent_dir = destination_path.parent
        dir_cmd = f"mkdir -p {parent_dir}"
        create_cmd = f"touch {destination}"
        write_command = f'cat <<"SCRIPTFILETAG" > {destination}'
        file_suffix = "SCRIPTFILETAG"
        cmds = [dir_cmd, create_cmd, write_command, file_contents, file_suffix]
        result = self.run_commands(cmds)
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to write file. stderr: {result.stderr}")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

from oumi.utils.logging import logger


class AliasType(str, Enum):
    """The type of configs we support with aliases."""

    TRAIN = "train"
    EVAL = "eval"
    INFER = "infer"
    JOB = "job"


_ALIASES: dict[str, dict[AliasType, str]] = {
    "llama4-scout": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_base_full/train.yaml",
        AliasType.JOB: "oumi://configs/recipes/llama4/sft/scout_base_full/train.yaml",
    },
    "llama4-scout-instruct-lora": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_lora/train.yaml",
    },
    "llama4-scout-instruct-qlora": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_qlora/train.yaml",
    },
    "llama4-scout-instruct": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_full/train.yaml",
        AliasType.INFER: "oumi://configs/recipes/llama4/inference/scout_instruct_infer.yaml",
        AliasType.JOB: "oumi://configs/recipes/llama4/sft/scout_instruct_full/gcp_job.yaml",
        AliasType.EVAL: "oumi://configs/recipes/llama4/evaluation/scout_instruct_eval.yaml",
    },
    "llama4-maverick": {
        AliasType.INFER: "oumi://configs/recipes/llama4/inference/maverick_instruct_together_infer.yaml",
    },
}


def try_get_config_name_for_alias(
    alias: str,
    alias_type: AliasType,
) -> str:
    """Gets the config path for a given alias.

    This function resolves the config path for a given alias and alias type.
    If the alias is not found, the original alias is returned.

    Args:
        alias (str): The alias to resolve.
        alias_type (AliasType): The type of config to resolve.

    Returns:
        str: The resolved config path (or the original alias if not found).
    """
    if alias in _ALIASES and alias_type in _ALIASES[alias]:
        config_path = _ALIASES[alias][alias_type]
        logger.info(f"Resolved alias '{alias}' to '{config_path}'")
        return config_path
    return alias



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Optional

import jsonlines
import typer
from rich.table import Table

from oumi.cli import cli_utils
from oumi.utils.io_utils import load_jsonlines

if TYPE_CHECKING:
    from oumi.core.configs import InferenceConfig, JudgeConfig


def _load_judge_config(config: str, extra_args: list[str]) -> "JudgeConfig":
    from oumi.core.registry import REGISTRY

    judge_config_builder = REGISTRY.get_judge_config(config)

    if judge_config_builder:
        if extra_args:
            typer.echo(
                "For consistent judge results, a named judge config cannot be "
                f"overridden with extra arguments. Got: {extra_args}. "
                "Please register a new named judge config, or provide a path to a "
                "judge config file."
            )
            raise typer.Exit(code=1)
        return judge_config_builder()

    if not Path(config).exists():
        typer.echo(f"Config file not found: '{config}'")
        raise typer.Exit(code=1)

    return JudgeConfig.from_yaml_and_arg_list(config, extra_args)


def dataset(
    ctx: typer.Context,
    config: Annotated[
        str, typer.Option(*cli_utils.CONFIG_FLAGS, help="Path to the judge config file")
    ],
    dataset_name: Annotated[
        Optional[str], typer.Option(help="Name of the dataset from the registry")
    ] = None,
    dataset_subset: Annotated[
        Optional[str], typer.Option(help="Subset of the dataset to use, if applicable")
    ] = None,
    dataset_split: Annotated[
        Optional[str], typer.Option(help="Split of the dataset to use.")
    ] = "train",
    output_file: Annotated[
        Optional[str], typer.Option(help="Path to the output file (jsonl)")
    ] = None,
    level: cli_utils.LOG_LEVEL_TYPE = None,
):
    """Judge a dataset."""
    # Delayed imports
    from oumi import judge_dataset
    from oumi.core.registry import REGISTRY

    # End imports
    if not dataset_name:
        typer.echo("Dataset name is required.")
        raise typer.Exit(code=1)

    # Load the judge config
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            config,
        )
    )

    judge_config = _load_judge_config(config, extra_args)

    # Load the dataset class from the registry
    dataset_class = REGISTRY.get_dataset(dataset_name, subset=dataset_subset)

    if dataset_class is None:
        typer.echo(f"Dataset '{dataset_name}' not found in registry.")
        raise typer.Exit(code=1)

    dataset = dataset_class(
        split=dataset_split,
        subset=dataset_subset,
    )

    # Judge the dataset
    results = judge_dataset(judge_config, dataset=dataset)

    # Save the results
    if output_file:
        typer.echo(f"Saving results to {output_file}")
        with jsonlines.open(output_file, mode="w") as writer:
            writer.write_all(results)
    else:
        table = Table(
            title="Judge Results",
            title_style="bold magenta",
            show_edge=False,
            show_lines=True,
        )
        table.add_column("Judgements", style="green")
        for result in results:
            table.add_row(json.dumps(result))
        cli_utils.CONSOLE.print(table)


def conversations(
    ctx: typer.Context,
    config: Annotated[
        str, typer.Option(*cli_utils.CONFIG_FLAGS, help="Path to the judge config file")
    ],
    input_file: Annotated[
        Optional[str], typer.Option(help="Path to the input file (jsonl)")
    ] = None,
    output_file: Annotated[
        Optional[str], typer.Option(help="Path to the output file (jsonl)")
    ] = None,
    level: cli_utils.LOG_LEVEL_TYPE = None,
):
    """Judge a list of conversations."""
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            config,
        )
    )

    # Delayed imports
    from oumi import judge_conversations
    from oumi.core.types.conversation import Conversation
    # End imports

    # Load the judge config
    judge_config = _load_judge_config(config, extra_args)

    # Load the conversations from the input file
    if not input_file:
        typer.echo("Input file is required.")
        raise typer.Exit(code=1)

    input_data = load_jsonlines(input_file)
    conversations = [Conversation.from_dict(conv) for conv in input_data]

    # Judge the conversations
    results = judge_conversations(judge_config, judge_inputs=conversations)

    # Save the results
    if output_file:
        typer.echo(f"Saving results to {output_file}")
        with jsonlines.open(output_file, mode="w") as writer:
            writer.write_all(results)
    else:
        table = Table(
            title="Judge Results",
            title_style="bold magenta",
            show_edge=False,
            show_lines=True,
        )
        table.add_column("Judgements", style="green")
        for result in results:
            table.add_row(json.dumps(result))
        cli_utils.CONSOLE.print(table)


def model(
    ctx: typer.Context,
    config: Annotated[
        str, typer.Option(*cli_utils.CONFIG_FLAGS, help="Path to the judge config file")
    ],
    inference_config: Annotated[
        str,
        typer.Option(help="Path to the inference config file"),
    ],
    input_file: Annotated[
        Optional[str], typer.Option(help="Path to the input file (jsonl)")
    ] = None,
    output_file: Annotated[
        Optional[str], typer.Option(help="Path to the output file (jsonl)")
    ] = None,
    level: cli_utils.LOG_LEVEL_TYPE = None,
):
    """Judge the outputs of a model on a dataset."""
    # Delayed imports
    from oumi import judge_conversations
    from oumi.builders.inference_engines import build_inference_engine
    from oumi.core.types.conversation import Conversation
    # End imports

    judge_extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            config,
        )
    )
    # Load the judge config
    judge_config = _load_judge_config(config, judge_extra_args)

    # Load the inference config
    inference_config = str(
        cli_utils.resolve_and_fetch_config(
            inference_config,
        )
    )
    inference_extra_args = cli_utils.parse_extra_cli_args(ctx)
    model_inference_config: InferenceConfig = InferenceConfig.from_yaml_and_arg_list(
        inference_config, inference_extra_args
    )

    if not model_inference_config.engine:
        typer.echo("Inference engine is required.")
        raise typer.Exit(code=1)

    # Load the dataset
    if not input_file:
        typer.echo("Input file is required.")
        raise typer.Exit(code=1)

    input_data = load_jsonlines(input_file)
    input_conversations = [Conversation.from_dict(output) for output in input_data]

    # Run inference
    inference_engine = build_inference_engine(
        model_inference_config.engine,
        model_params=model_inference_config.model,
        remote_params=model_inference_config.remote_params,
        generation_params=model_inference_config.generation,
    )

    model_outputs = inference_engine.infer(
        input=input_conversations, inference_config=model_inference_config
    )

    results = judge_conversations(judge_config, judge_inputs=model_outputs)

    if output_file:
        typer.echo(f"Saving results to {output_file}")
        with jsonlines.open(output_file, mode="w") as writer:
            writer.write_all(results)
    else:
        table = Table(
            title="Judge Results",
            title_style="bold magenta",
            show_edge=False,
            show_lines=True,
        )
        table.add_column("Judgements", style="green")
        for result in results:
            table.add_row(json.dumps(result))
        cli_utils.CONSOLE.print(table)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import jsonlines
from tqdm import tqdm

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
)
from oumi.core.types.conversation import Conversation
from oumi.utils.logging import logger


class BaseInferenceEngine(ABC):
    """Base class for running model inference."""

    _model_params: ModelParams
    """The model parameters."""

    _generation_params: GenerationParams
    """The generation parameters."""

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: Optional[GenerationParams] = None,
    ):
        """Initializes the inference engine.

        Args:
            model_params: The model parameters.
            generation_params: The generation parameters.
        """
        self._model_params = copy.deepcopy(model_params)
        if generation_params:
            self._check_unsupported_params(generation_params)
        else:
            generation_params = GenerationParams()
        self._generation_params = generation_params

    def infer(
        self,
        input: Optional[list[Conversation]] = None,
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference.

        Args:
            input: A list of conversations to run inference on. Optional.
            inference_config: Parameters for inference.
                If not specified, a default config is inferred.

        Returns:
            List[Conversation]: Inference output.
        """
        if input is not None and (
            inference_config and inference_config.input_path is not None
        ):
            raise ValueError(
                "Only one of input or inference_config.input_path should be provided."
            )

        # Ensure the inference config has up-to-date generation parameters.
        if inference_config:
            if inference_config.generation:
                self._check_unsupported_params(inference_config.generation)
            elif self._generation_params:
                inference_config = copy.deepcopy(inference_config)
                inference_config.generation = self._generation_params

                # Warn the user: They provided an inference config without generation
                # params, so what was the point of providing it in the first place?
                logger.warning(
                    "No generation parameters provided in the inference config. Using "
                    "the generation parameters that the engine was initialized with."
                )

        if input is not None:
            return self.infer_online(input, inference_config)
        elif inference_config and inference_config.input_path is not None:
            return self.infer_from_file(inference_config.input_path, inference_config)
        else:
            raise ValueError(
                "One of input or inference_config.input_path must be provided."
            )

    def _read_conversations(self, input_filepath: str) -> list[Conversation]:
        """Reads conversations from a file in Oumi chat format.

        Args:
            input_filepath: The path to the file containing the conversations.

        Returns:
            List[Conversation]: A list of conversations read from the file.
        """
        conversations = []
        with open(input_filepath) as f:
            for line in f:
                # Only parse non-empty lines.
                if line.strip():
                    conversation = Conversation.from_json(line)
                    conversations.append(conversation)
        return conversations

    def _get_scratch_filepath(self, output_filepath: str) -> str:
        """Returns a scratch filepath for the given output filepath.

        For example, if the output filepath is "/foo/bar/output.json", the scratch
        filepath will be "/foo/bar/scratch/output.json"

        Args:
            output_filepath: The output filepath.

        Returns:
            str: The scratch filepath.
        """
        original_filepath = Path(output_filepath)
        return str(original_filepath.parent / "scratch" / original_filepath.name)

    def _save_conversation(
        self, conversation: Conversation, output_filepath: str
    ) -> None:
        """Appends a conversation to a file in Oumi chat format.

        Args:
            conversation: The conversation to save.
            output_filepath: The path to the file where the conversation should be
                saved.
        """
        # Make the directory if it doesn't exist.
        Path(output_filepath).parent.mkdir(parents=True, exist_ok=True)
        with jsonlines.open(output_filepath, mode="a") as writer:
            json_obj = conversation.to_dict()
            writer.write(json_obj)

    def _save_conversations(
        self, conversations: list[Conversation], output_filepath: str
    ) -> None:
        """Saves conversations to a file in Oumi chat format.

        Args:
            conversations: A list of conversations to save.
            output_filepath: The path to the file where the conversations should be
                saved.
        """
        # Make the directory if it doesn't exist.
        Path(output_filepath).parent.mkdir(parents=True, exist_ok=True)
        with jsonlines.open(output_filepath, mode="w") as writer:
            for conversation in tqdm(conversations, desc="Saving conversations"):
                json_obj = conversation.to_dict()
                writer.write(json_obj)

    def _check_unsupported_params(self, generation_params: GenerationParams):
        """Checks for unsupported parameters and logs warnings.

        If a parameter is not supported, and a non-default value is provided,
        a warning is logged.
        """
        supported_params = self.get_supported_params()
        default_generation_params = GenerationParams()

        for param_name, value in generation_params:
            if param_name not in supported_params:
                is_non_default_value = (
                    getattr(default_generation_params, param_name) != value
                )

                if is_non_default_value:
                    logger.warning(
                        f"{self.__class__.__name__} does not support {param_name}. "
                        f"Received value: {param_name}={value}. "
                        "This parameter will be ignored."
                    )

    @abstractmethod
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine.

        Override this method in derived classes to specify which parameters
        are supported.

        Returns:
            Set[str]: A set of supported parameter names.
        """
        raise NotImplementedError

    @abstractmethod
    def infer_online(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        raise NotImplementedError

    @abstractmethod
    def infer_from_file(
        self,
        input_filepath: str,
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference on inputs in the provided file.

        This is a convenience method to prevent boilerplate from asserting the existence
        of input_filepath in the generation_params.

        Args:
            input_filepath: Path to the input file containing prompts for generation.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        raise NotImplementedError

    def apply_chat_template(
        self, conversation: Conversation, **tokenizer_kwargs
    ) -> str:
        """Applies the chat template to the conversation.

        Args:
            conversation: The conversation to apply the chat template to.
            tokenizer_kwargs: Additional keyword arguments to pass to the tokenizer.

        Returns:
            str: The conversation with the chat template applied.
        """
        tokenizer = getattr(self, "_tokenizer", None)

        if tokenizer is None:
            raise ValueError("Tokenizer is not initialized.")

        if tokenizer.chat_template is None:
            raise ValueError("Tokenizer does not have a chat template.")

        if "tokenize" not in tokenizer_kwargs:
            tokenizer_kwargs["tokenize"] = False

        return tokenizer.apply_chat_template(conversation, **tokenizer_kwargs)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from typing_extensions import override

from oumi.core.datasets.base_map_dataset import BaseMapDataset
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer

_PROMPT_KEY = "prompt"
_CHOSEN_KEY = "chosen"
_REJECTED_KEY = "rejected"

_ROLE = "role"
_CONTENT = "content"
_ASSISTANT = "assistant"


class BaseExperimentalDpoDataset(BaseMapDataset):
    """Preprocess the samples to the Oumi format.

    Warning:
        This class is experimental and subject to change.
    """

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[str] = None,
        split: Optional[str] = None,
        tokenizer: Optional[BaseTokenizer] = None,
        return_tensors: bool = False,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the BaseExperimentalDpoDataset class."""
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=dataset_path,
            split=split,
            **kwargs,
        )

        if return_tensors:
            raise NotImplementedError(
                "return_tensors=True is not implemented for this class"
            )

        self._tokenizer = tokenizer
        self._return_tensors = return_tensors

        self._data = self._load_data()

    def transform_preference(self, samples: dict) -> dict:
        """Transform the samples to the Oumi format."""
        prompt = samples[_PROMPT_KEY]
        chosen_chat = samples[_CHOSEN_KEY]
        rejected_chat = samples[_REJECTED_KEY]

        chosen_chat_response = self._extract_from_chat_format(chosen_chat)
        rejected_chat_response = self._extract_from_chat_format(rejected_chat)

        return {
            _PROMPT_KEY: prompt,
            _CHOSEN_KEY: chosen_chat_response,
            _REJECTED_KEY: rejected_chat_response,
        }

    @override
    def transform(self, sample: dict) -> dict:
        """Transform the samples to the Oumi format."""
        return self.transform_preference(sample)

    def _extract_from_chat_format(self, sample: dict) -> str:
        """Extract the last 'assistant' turn in the chat."""
        for turn in sample[::-1]:
            if turn[_ROLE] == _ASSISTANT:
                return turn[_CONTENT]

        raise ValueError("No chat turn was found with an 'assistant' role.")
